/*
 * Developed by Hillstone Software. http://www.hillstone-software.com
 * SMTP library API definitions
 */
#ifndef _HSSMTP_H
#define _HSSMTP_H

#define HS_SMTP_MAX_DOMAIN          65                          // max domain name (including terminating 0)

typedef int smtp_callback_t(void *session_ref, int ev, long arg1, long arg2);

typedef struct
{
	/* either name or IP address may be specified, if not specified supply NULL pointer */
	/* if srv_name is NULL, try to connect using srv_ip - which is a string in the form x.x.x.x */
	/* if srv_name is not null, HS SMTP will try to resolve name to IP address */	
	unsigned char	*srv_name;		// SMTP server name		
	unsigned char	*srv_ip;		// SMTP server ip address
	
	unsigned short	srv_port;		// SMTP server port
	
	unsigned char	*send_addr;		// sender email address
	unsigned char	*recv_addr;		// receiver email address
	
	unsigned char	*subj;			// subject

	int				subject_charset; // character set to be used in the subject field
	
	unsigned char 	*cc;			// cc email address
	
	unsigned char	*msgp; 			// pointer to message text buffer
	int				msglen;			// message text length

    int             do_auth;        // Do Authentication
    unsigned char   *usrname;       // user name for authentication
    unsigned char   *password;      // password for authentication

	smtp_callback_t	*callback;		// callback function used by HS SMTP to report completion status
	void			*session_ref;	// user reference association with this session

	int				send_file;		// 1=send file attachment, 0=no file attachment
	unsigned char	*filename;		// filename of file to send

	int				use_latin1;		// for characters from 0x7f to 0xff use escape sequence ==<hex code> and use .charset="iso-8859-1"
} hs_smtp_mail_t;

typedef enum
{
	HS_SMTP_CHARSET_DEFAULT,		// default charset
	HS_SMTP_CHARSET_KOI8R,			// KOI8-R
} hs_smtp_charset_t;

/* return codes */
typedef enum
{
	HS_SMTP_RC_OK,					// success
	HS_SMTP_RC_ALRINIT,				// already initialised
	HS_SMTP_RC_NOTINIT,				// not initialised
	HS_SMPT_RC_SOCKINIT_FAIL,		// socket layer initialisation failed
	HS_SMPT_RC_SOCKINIT2_FAIL,      // socket layer failure (could not get local host name)
	HS_SMTP_RC_INV_PAR,				// invalid papameters
	HS_SMTP_RC_NONAME,				// can't resolve hostname
	HS_SMTP_RC_NO_CTX,				// no free session contexts left
	HS_SMTP_RC_TCPCONNFAIL,         // failed to connect to server
	HS_SMTP_RC_NOT_OPEN,			// Session is not open
	HS_SMTP_RC_FOPEN_ERR,			// file open error
	HS_SMTP_RC_NO_MEM,				// no free memory

	HS_SMTP_RC_NUM,
} hs_smtp_rc_t;

/* callback events */
typedef enum
{
	HS_SMTP_USR_EV_DONE=1,			// Mail Transation Complete, mail has been successfully sent to SMTP server
	HS_SMTP_USR_EV_CLOSED,			// Socket layer closed TCP connection
	HS_SMTP_USR_EV_CONNFAIL,		// TCP connection request failed
	HS_SMTP_USR_EV_SRVERR,			// Session closed because of SMTP error reply received from server
	HS_SMTP_USR_EV_TIMEDOUT,		// Timed out waiting for server reply
	HS_SMTP_USR_EV_PROGRESS,		// HS SMTP has transmitted a number of bytes of mail data
	HS_SMTP_USR_EV_PROGRESS_F,		// HS SMTP has transmitted a number of bytes of file data
	HS_SMTP_USR_EV_GETNEXT,         // HS SMTP Requests next message to send 
} hs_smtp_usr_ev_t;

typedef void hs_smtp_timer_cb_t(void *arg);
typedef long hs_smtp_timer_start_t(unsigned long timeout_ms, void *arg, hs_smtp_timer_cb_t *cb);
typedef void hs_smtp_timer_stop_t(long timer_han);

typedef void *hssmtp_create_file_t(unsigned char *filename, int mode, int open_always);
typedef void hssmtp_close_file_t(void *hFile);
typedef int hssmtp_write_file_t(void *hFile, void *pData, int length, int *bytes_written);
typedef int hssmtp_read_file_t(void *hFile, void *pData, int length, int *bytes_read);
typedef void hssmtp_get_file_size_t(void *hFile, long *fsize);

typedef enum
{
	HSSMTP_FILE_READ,
	HSSMTP_FILE_WRITE,
} hssmtp_file_mode_t;

// callback used by HSSMTP to get number of millisecond ticks since bootup
// Used for generating pseudo random initial sequence number
// if callback pointer not initialized, HSTCP will use default 
// pseudo random number generator to produce initial sequence numbers
typedef unsigned long hssmtp_get_ticks_fn_t(void);


typedef struct 
{
	unsigned char domain[HS_SMTP_MAX_DOMAIN];	// local domain name
	hs_smtp_timer_start_t	*start_timer;		// start timer callback
	hs_smtp_timer_stop_t	*stop_timer;		// stop timer callback
	hssmtp_create_file_t	*create_file;		// create file callback
	hssmtp_close_file_t		*close_file;		// close file callback
	hssmtp_write_file_t		*write_file;		// write file callback
	hssmtp_read_file_t		*read_file;			// read file callback
	hssmtp_get_file_size_t	*getfile_size;		// get file size callback
	hssmtp_get_ticks_fn_t	*getticks;			// get tick count callback
} hssmtp_init_t;

/*
 * Initialise HS SMTP Library
 */
extern int	HsSmtpInit(hssmtp_init_t *init);

/*
 * Release all HS SMTP resources and clean up
 */
extern int	HsSmtpDestroy(void);

/* HS SMTP tick function, drives internal time based operations, 
   must be called from user code as often as possible */
extern int	HsSmtpTick(void);

/*
 * Send Mail (with or without file attachment)
 */
extern int	HsSmtpSendMail(hs_smtp_mail_t	*m, long *session_handle);

/*
 * Abort mail transaction and close connectiuon to server
 */
extern int	HsSmtpAbortMail(long session_handle);

extern void  HsSmtpErrStr(unsigned char *str2, int rc);

#endif /* _HSSMTP_H */
