/*
 * HS FTP Client Library
 * Recursive folder related operations
 * Developed by Hillstone Software
 * http://www.hillstone-software.com
 * support@hillstone-software.com
 */
#include "HsFtpRecurs.h"

/* main control structure */
static hsftprecurs_t ctrl = {0};

static int get_file_size(HANDLE hFile, __int64 *fsize)
{
	DWORD dwSizeLow, dwSizeHigh, dwError;
	
    dwSizeLow = GetFileSize (hFile, &dwSizeHigh);          
    if (dwSizeLow == 0xFFFFFFFF && 
        (dwError = GetLastError()) != NO_ERROR )
    { 
		return FALSE;    	
    }     

	*fsize = (__int64)((dwSizeHigh * MAXDWORD) + dwSizeLow); 

	return TRUE;
}

/* check if file exists and if yes returns its file size */
static int file_exists(unsigned char *fname, __int64 *fsize)
{
	HANDLE hFile;

	hFile = CreateFile(fname, GENERIC_READ, 0, NULL, OPEN_EXISTING, FILE_ATTRIBUTE_NORMAL, NULL);
	if(hFile == INVALID_HANDLE_VALUE)
		return FALSE;

	get_file_size(hFile, fsize);

	CloseHandle(hFile);
	return TRUE;
}

static int hsftp_recurs_qput(recurs_q_t *q, long session_handle, int ev, int core_ev, unsigned char *pBuf, int len)
{
	recurs_evt_q_cell_t *pCell;

	pCell = &q->cells[q->in];
	memset(pCell, 0, sizeof (recurs_evt_q_cell_t));
	
	pCell->ev = ev;
	pCell->core_ev = core_ev;
	pCell->session_handle = session_handle;

	if ((pBuf) && (len))
	{
		pCell->pBuf = (unsigned char *)malloc(len);
		if (!pCell->pBuf) return FALSE;

		memcpy(pCell->pBuf, pBuf, len);
		pCell->len = len;
	}

	q->in++;

	if (q->in == HSFTP_RECURS_MAXQ)
		q->in = 0;

	return TRUE;
}

static int hsftp_recurs_qget(recurs_q_t *q, long *session_handle, int *ev, int *core_ev, unsigned char **ppBuf, int *len)
{
	recurs_evt_q_cell_t *pCell;
	
	if (q->in == q->out)
		return FALSE;

	pCell = &q->cells[q->out];

	*ev = pCell->ev;
	*core_ev = pCell->core_ev;
	*session_handle = pCell->session_handle;

	*len = 0;
	(*ppBuf) = NULL;

	if ((ppBuf) && (pCell->pBuf) && (pCell->len))
	{
		(*ppBuf) = (unsigned char *)malloc(pCell->len);
		if ((*ppBuf) != NULL)
		{		
			*len = pCell->len;
			if (pCell->len)
				memcpy((*ppBuf), pCell->pBuf, pCell->len);
		}
	}

	if (pCell->pBuf) free(pCell->pBuf);
	pCell->pBuf = NULL;
	pCell->len = 0;

	q->out++;
	if (q->out == HSFTP_RECURS_MAXQ)
		q->out = 0;

	return TRUE;
	
}

/* change state */
static void hsftp_recurs_change_state(hs_ftp_recurs_fsm_t *newstate)
{
	ctrl.fsm = newstate;
}

/* pass completion of failure event to user */
static void hsftp_recurs_event_to_user(long ftp_session, int ev, long arg1, long arg2)
{
	if (ctrl.cb) 
		ctrl.cb(ftp_session, ev, arg1, arg2);
}

/* find substring number n in string src, return in *index
 * string must be separated by spaces */
static int find_string_start_n(unsigned char *src, int n, int *index)
{
    int i, k;
    int is_str = FALSE;
    int len = strlen(src);

    k = 0;
    for (i=0; i<len; i++)
    {
        // if not within string
        if (is_str == FALSE)
        {
            if (src[i] != 0x20)
            {
                k++;
                if (k == n)
                {
                    *index = i;
                    return TRUE;
                }
                is_str = TRUE;
            }            
        }
        // withing string
        else
        {
            if (src[i] == 0x20)
            {
                is_str = FALSE;
            }
        }
    }

    return FALSE;
}

static int find_string_byte(unsigned char *src, unsigned char ch, int *i)
{
	int n = 0;

	while (*src++)
	{
		if ((*src) == ch)
		{
			*i = (n+1);
			return TRUE;
		}
		n++;
	}
	return FALSE;
}

/* recursive folder delete failed */
static void hsftp_recurs_on_folder_delete_failed(long session_handle, int rc, int core_ev)
{
	ListFree(&ctrl.file_tree);

	hsftp_recurs_change_state(hsftp_recurs_state_idle);
	
	hsftp_recurs_event_to_user(session_handle, HSFTP_RECURS_USR_EV_FOLDER_DELETE_FAILED, rc, core_ev);
}

/* recursive folder download failed */
static void hsftp_recurs_on_folder_download_failed(long session_handle, int rc, int core_ev)
{
	ListFree(&ctrl.file_tree);

	hsftp_recurs_change_state(hsftp_recurs_state_idle);

	chdir(ctrl.start_dir);
	
	hsftp_recurs_event_to_user(session_handle, HSFTP_RECURS_USR_EV_FOLDER_DOWNLOAD_FAILED, rc, core_ev);	
}

/* recursive folder upload failed */
static void hsftp_recurs_on_folder_upload_failed(long session_handle, int rc, int core_ev)
{
	ListFree(&ctrl.file_tree);

	hsftp_recurs_change_state(hsftp_recurs_state_idle);

	chdir(ctrl.start_dir);
	
	hsftp_recurs_event_to_user(session_handle, HSFTP_RECURS_USR_EV_FOLDER_UPLOAD_FAILED, rc, core_ev);	
}

/* start recursive folder delete */
static void hsftp_recurs_start_folder_delete(long session_handle, unsigned char *foldername)
{
	list_t				*pChild;
	list_item_t			*p;
	int					rc;
	long				errcode = HSFTP_RECURS_RC_NOMEM;

	ListFree(&ctrl.file_tree);
	ListInit(&ctrl.file_tree);

	pChild = (list_t *)malloc(sizeof (list_t));
	if (pChild)
	{
		ListInit(pChild);

		p = ListPut(&ctrl.file_tree, foldername, strlen(foldername) + 1);
		if (p)
		{
			p->list = pChild;
			ctrl.pCurList = pChild;
			pChild->pParent = &ctrl.file_tree;

			rc = HsFtpCliChDir(session_handle, foldername);
			if (rc == HS_FTP_RC_OK)
			{
				hsftp_recurs_change_state(hsftp_recurs_state_fdel_chdir_down);
				return;
			}
			errcode = HSFTP_RECURS_RC_RCHDIR; 		
		}
	}

	hsftp_recurs_on_folder_delete_failed(session_handle, errcode, 0);	
}

/* start recursive folder download */
static void hsftp_recurs_start_folder_download(long session_handle, unsigned char *foldername)
{
	list_t				*pChild;
	list_item_t			*p;
	int					rc;

	CreateDirectory(foldername, NULL);

	if (chdir(foldername) != 0)
	{
		hsftp_recurs_on_folder_download_failed(session_handle, HSFTP_RECURS_RC_LCHDIR, 0);
		return;
	}

	ListFree(&ctrl.file_tree);
	ListInit(&ctrl.file_tree);	
		
	pChild = (list_t *)malloc(sizeof (list_t));
	if (!pChild)
	{
		hsftp_recurs_on_folder_download_failed(session_handle, HSFTP_RECURS_RC_NOMEM, 0);
		return;
	}
	
	ListInit(pChild);

	p = ListPut(&ctrl.file_tree, foldername, strlen(foldername) + 1);
	if (!p)
	{
		hsftp_recurs_on_folder_download_failed(session_handle, HSFTP_RECURS_RC_NOMEM, 0);
		return;
	}
		
	p->list = pChild;
	ctrl.pCurList = pChild;
	pChild->pParent = &ctrl.file_tree;

	rc = HsFtpCliChDir(session_handle, foldername);
	if (rc != HS_FTP_RC_OK)
	{
		hsftp_recurs_on_folder_download_failed(session_handle, HSFTP_RECURS_RC_RCHDIR, 0);
		return;
	}

	hsftp_recurs_change_state(hsftp_recurs_state_fget_chdir_down);
}

/* start recursive folder upload */
static void hsftp_recurs_start_folder_upload(long session_handle, unsigned char *foldername)
{
	list_t				*pChild;
	list_item_t			*p;
	int					rc;

	ListFree(&ctrl.file_tree);
	ListInit(&ctrl.file_tree);	
		
	pChild = (list_t *)malloc(sizeof (list_t));
	if (!pChild)
	{
		hsftp_recurs_on_folder_upload_failed(session_handle, HSFTP_RECURS_RC_NOMEM, 0);
		return;
	}
	
	ListInit(pChild);

	p = ListPut(&ctrl.file_tree, foldername, strlen(foldername) + 1);
	if (!p)
	{
		hsftp_recurs_on_folder_upload_failed(session_handle, HSFTP_RECURS_RC_NOMEM, 0);
		return;
	}
		
	p->list = pChild;
	ctrl.pCurList = pChild;
	pChild->pParent = &ctrl.file_tree;

	rc = HsFtpCliCreateDir(session_handle, foldername);
	if (rc != HS_FTP_RC_OK)
	{
		hsftp_recurs_on_folder_upload_failed(session_handle, HSFTP_RECURS_RC_RCHDIR, 0);
		return;
	}

	hsftp_recurs_change_state(hsftp_recurs_state_fput_mkdir);
}

/* Process next file entry during recursive folder delete */
static int hsftp_recurs_fdelete_get_next_file(long session_handle)
{
	list_item_t			*p;
	list_t				*pChild;
	list_t				*pThisList;
	int					rc;

	pThisList = ctrl.pCurList; 
	p = ctrl.pCurList->pMarker;
	
	// if current item is '.' or '..', skip to next item
	if ((p) && 
		((strcmp((unsigned char *)p->data, ".") == 0) ||
		(strcmp((unsigned char *)p->data, "..") == 0)))
		p = (list_item_t *)p->next;

	// do it again, in case there are two entries like '.' and '..'
	if ((p) && 
		((strcmp((unsigned char *)p->data, ".") == 0) ||
		(strcmp((unsigned char *)p->data, "..") == 0)))
		p = (list_item_t *)p->next;

	// if we are at the end of the list (no more files to get)
	if (!p)
	{		
		// move back up one level
		ctrl.pCurList = ctrl.pCurList->pParent;

		hsftp_recurs_change_state(hsftp_recurs_state_fdel_chdir_up);
			
		rc = HsFtpCliChDir(session_handle, "..");
		if (rc != HS_FTP_RC_OK)
		{
			hsftp_recurs_on_folder_delete_failed(session_handle, HSFTP_RECURS_RC_RCHDIR, 0);
			return FALSE;
		}

		return TRUE; 
	}

	// if current item is a directory
	if (*((unsigned char *)p->data_ex2) == 'd')
	{		
		pChild = (list_t *)malloc(sizeof (list_t));
		if (!pChild)
		{
			hsftp_recurs_on_folder_delete_failed(session_handle, HSFTP_RECURS_RC_NOMEM, 0);
			return FALSE;	
		}

		ListInit(pChild);

		p->list = pChild;
		pChild->pParent = ctrl.pCurList;
		ctrl.pCurList = pChild;
		
		hsftp_recurs_change_state(hsftp_recurs_state_fdel_chdir_down);

		rc = HsFtpCliChDir(session_handle, (unsigned char *)p->data);
		if (rc != HS_FTP_RC_OK)
		{
			hsftp_recurs_on_folder_delete_failed(session_handle, HSFTP_RECURS_RC_RCHDIR, 0);
			return FALSE;
		}

		pThisList->pMarker = (list_item_t *)p->next;
		return TRUE;
	}

	hsftp_recurs_event_to_user(
		session_handle, HSFTP_RECURS_USR_EV_FOLDER_DELETE_STATUS, 
		(long)p->data, HSFTP_OP_START);
	
	hsftp_recurs_change_state(hsftp_recurs_state_fdel_file);

	rc = HsFtpCliDeleteFile(session_handle, (unsigned char *)p->data);
	if (rc != HS_FTP_RC_OK)
	{
		hsftp_recurs_on_folder_delete_failed(session_handle, HSFTP_RECURS_RC_FILE, 0);
		return FALSE;
	}	

	pThisList->pMarker = (list_item_t *)p->next;
	return TRUE;
	
}

/* Process next file entry during recursive folder download */
static int hsftp_recurs_fdownload_get_next_file(long session_handle)
{
	list_item_t			*p;
	list_t				*pChild;
	list_t				*pThisList;
	int					rc;
	__int64				local_file_size;

move_to_next_item:
	
	pThisList = ctrl.pCurList; 
	p = ctrl.pCurList->pMarker;

	// if current item is '.' or '..', skip to next item
	if ((p) && 
		((strcmp((unsigned char *)p->data, ".") == 0) ||
		(strcmp((unsigned char *)p->data, "..") == 0)))
		p = (list_item_t *)p->next;

	// do it again, in case there are two entries like '.' and '..'
	if ((p) && 
		((strcmp((unsigned char *)p->data, ".") == 0) ||
		(strcmp((unsigned char *)p->data, "..") == 0)))
		p = (list_item_t *)p->next;

	// if we are at the end of the list (no more files to get)
	if (!p)
	{		
		// move back up one level
		ctrl.pCurList = ctrl.pCurList->pParent;

		chdir("..");

		// if returned to root
		if (ctrl.pCurList == &ctrl.file_tree)
			hsftp_recurs_change_state(hsftp_recurs_state_fget_chdir_root);
		else // go one level up
			hsftp_recurs_change_state(hsftp_recurs_state_fget_chdir_up);

		rc = HsFtpCliChDir(session_handle, "..");
		if (rc != HS_FTP_RC_OK)
		{
			hsftp_recurs_on_folder_download_failed(session_handle, HSFTP_RECURS_RC_RCHDIR, 0);
			return FALSE;
		}
		
		return TRUE; 
	}

	// if current item is a directory
	if (*((unsigned char *)p->data_ex2) == 'd')
	{		
		pChild = (list_t *)malloc(sizeof (list_t));
		if (!pChild)
		{
			hsftp_recurs_on_folder_download_failed(session_handle, HSFTP_RECURS_RC_NOMEM, 0);
			return FALSE;	
		}

		ListInit(pChild);

		p->list = pChild;
		pChild->pParent = ctrl.pCurList;
		ctrl.pCurList = pChild;

		CreateDirectory((unsigned char *)p->data, NULL);
		chdir((unsigned char *)p->data);
		
		hsftp_recurs_change_state(hsftp_recurs_state_fget_chdir_down);

		rc = HsFtpCliChDir(session_handle, (unsigned char *)p->data);
		if (rc != HS_FTP_RC_OK)
		{
			hsftp_recurs_on_folder_download_failed(session_handle, HSFTP_RECURS_RC_RCHDIR, 0);
			return FALSE;
		}

		pThisList->pMarker = (list_item_t *)p->next;
		return TRUE;
	}

	ctrl.fsize = _atoi64((unsigned char *)p->data_ex1);

	if ((!ctrl.overwrite) && (file_exists((unsigned char *)p->data, &local_file_size)))
	{
		if (local_file_size == ctrl.fsize)
		{
			pThisList->pMarker = (list_item_t *)p->next;
			goto move_to_next_item;
		}
	}

	hsftp_recurs_event_to_user(
		session_handle, HSFTP_RECURS_USR_EV_FOLDER_DOWNLOAD_STATUS, 
		(long)p->data, HSFTP_OP_START);
	
	hsftp_recurs_change_state(hsftp_recurs_state_fget_file);

	hsftp_recurs_event_to_user(
		session_handle, HSFTP_RECURS_USR_EV_FOLDER_PROGRESS, 0, 0);

	ctrl.rxed_len = 0;

	rc = HsFtpCliGetFile(session_handle, (unsigned char *)p->data, ctrl.fsize);
	if (rc != HS_FTP_RC_OK)
	{
		hsftp_recurs_on_folder_download_failed(session_handle, HSFTP_RECURS_RC_FILE, 0);
		return FALSE;
	}	

	pThisList->pMarker = (list_item_t *)p->next;
	return TRUE;
}

/* Process (start upload) next file from the current level listing */
extern int hsftp_recurs_fupload_get_next_file(long session_handle)
{
	list_item_t			*p;
	list_t				*pChild;
	list_t				*pThisList;
	int					rc;
	int					length;
	
	pThisList = ctrl.pCurList; 
	p = ctrl.pCurList->pMarker;

	// if current item is '.' or '..', skip to next item
	if ((p) && 
		((strcmp((unsigned char *)p->data, ".") == 0) ||
		(strcmp((unsigned char *)p->data, "..") == 0)))
		p = (list_item_t *)p->next;

	// do it again, in case there are two entries like '.' and '..'
	if ((p) && 
		((strcmp((unsigned char *)p->data, ".") == 0) ||
		(strcmp((unsigned char *)p->data, "..") == 0)))
		p = (list_item_t *)p->next;

	// if we are at the end of the list (no more files to get)
	if (!p)
	{
		// move back up one level
		ctrl.pCurList = ctrl.pCurList->pParent;

		chdir("..");

		// if returned to root
		if (ctrl.pCurList == &ctrl.file_tree)
			hsftp_recurs_change_state(hsftp_recurs_state_fput_chdir_root);
		else // go one level up
			hsftp_recurs_change_state(hsftp_recurs_state_fput_chdir_up);

		rc = HsFtpCliChDir(session_handle, "..");
		if (rc != HS_FTP_RC_OK)
		{
			hsftp_recurs_on_folder_upload_failed(session_handle, HSFTP_RECURS_RC_RCHDIR, 0);
			return FALSE;
		}

		return TRUE; 
	}

	// if current item is a directory
	if (*((unsigned char *)p->data_ex2) == 'd')
	{		
		pChild = (list_t *)malloc(sizeof (list_t));
		if (!pChild)
		{
			hsftp_recurs_on_folder_upload_failed(session_handle, HSFTP_RECURS_RC_NOMEM, 0);
			return FALSE;	
		}

		ListInit(pChild);

		p->list = pChild;
		pChild->pParent = ctrl.pCurList;
		ctrl.pCurList = pChild;
		
		chdir((unsigned char *)p->data);

		length = (int)strlen((unsigned char *)p->data);
		if (length >= sizeof(ctrl.path))
		{
			hsftp_recurs_on_folder_upload_failed(session_handle, HSFTP_RECURS_RC_PATH, 0);
			return FALSE;			
		}
		
		memset(ctrl.path, 0, sizeof(ctrl.path));
		memcpy(ctrl.path, (unsigned char *)p->data, length);

		hsftp_recurs_change_state(hsftp_recurs_state_fput_mkdir);

		rc = HsFtpCliCreateDir(session_handle, (unsigned char *)p->data);
		if (rc != HS_FTP_RC_OK)
		{
			hsftp_recurs_on_folder_upload_failed(session_handle, HSFTP_RECURS_RC_RMKDIR, 0);
			return FALSE;
		}

		pThisList->pMarker = (list_item_t *)p->next;

		return TRUE;
	}

	ctrl.txed_len = 0;
	ctrl.fsize = _atoi64((unsigned char *)p->data_ex1);

	hsftp_recurs_event_to_user(
		session_handle, HSFTP_RECURS_USR_EV_FOLDER_UPLOAD_STATUS, 
		(long)p->data, HSFTP_OP_START);
	
	hsftp_recurs_change_state(hsftp_recurs_state_fput_file);

	hsftp_recurs_event_to_user(
		session_handle, HSFTP_RECURS_USR_EV_FOLDER_PROGRESS, 0, 0);

	rc = HsFtpCliSendFile(session_handle, (unsigned char *)p->data);
	if (rc != HS_FTP_RC_OK)
	{
		hsftp_recurs_on_folder_upload_failed(session_handle, HSFTP_RECURS_RC_FSEND, 0);
		return FALSE;
	}	

	pThisList->pMarker = (list_item_t *)p->next;

	return TRUE;
}

/* deleting folder: process listing after having changed into remote directory */
/* downloading folder: process listing after having changed into remote directory */
static void hsftp_recurs_fdel_process_listing_chdir_down(long session_handle, unsigned char *buf, int len)
{
    int                 i,n;
    unsigned char       line[256] = {0};
    unsigned char		s[80];
    unsigned char		s1[2]; 
    unsigned char       ch;
    int                 index, index1, k;
    int                 fn_num;
    int					sz_num;
    int					is_unix;
	list_item_t			*p;

	if (!buf)
	{		
		p = ListPut(ctrl.pCurList, "..", 3);
		if (!p) return;

		ItemAddDataEx1(p, "0", 2);
		ItemAddDataEx2(p, "d", 2);

		ctrl.pCurList->pMarker = ctrl.pCurList->head;

		if (ctrl.fsm == hsftp_recurs_state_fdel_down_waitlist)
			hsftp_recurs_fdelete_get_next_file(session_handle);
		else if (ctrl.fsm == hsftp_recurs_state_fget_waitlist)
			hsftp_recurs_fdownload_get_next_file(session_handle);
		
		return;	
	}

    i = 0; n = 0;
    while (i < len)
    {
        ch = buf[i];
        if (ch == 10) 
            ;
        else if (ch == 13)
        {        
            if (strnicmp(line, "total", 5) != 0)
            {
                // if unix, first byte is non digit.
				is_unix = (isdigit(line[0])) ? 0 : 1;

				memset(s1, 0, sizeof (s1));

				if (is_unix)
				{
					fn_num = 9;
					sz_num = 5;
					s1[0] = line[0];
				}
				else
				{
					fn_num = 4;
					sz_num = 3;
					s1[0] = ' ';
				}

                // filename
                if (!find_string_start_n(line, fn_num, &index))
                    break;

				// size
				if (!find_string_start_n(line, sz_num, &index1))
					break;

				if (!find_string_byte(&line[index1], 0x20, &k))
					break;

				if (k > (sizeof(s) - 1))
					break;
					
				memset(s, 0, sizeof (s));
				memcpy(s, &line[index1], k);

				// if "<DIR>" is found in place of size, we are dealing with directory
				if (stricmp(s, "<DIR>") == 0)
				{
					s[0] = '0'; s[1] = 0;
					s1[0] = 'd'; s1[1] = 0;	
				}
				
				p = ListPut(ctrl.pCurList, &line[index], strlen(&line[index]) + 1);
				if (!p) return;

				ItemAddDataEx1(p, s, strlen(s)+1);
				ItemAddDataEx2(p, s1, strlen(s1)+1);
            }
            
            n = 0;
            memset(line, 0, sizeof(line));
        }
        else
        {
            line[n++] = ch;
        }

        i++;
    }  

	ctrl.pCurList->pMarker = ctrl.pCurList->head;

	if (ctrl.fsm == hsftp_recurs_state_fdel_down_waitlist)
		hsftp_recurs_fdelete_get_next_file(session_handle);
	else if (ctrl.fsm == hsftp_recurs_state_fget_waitlist)
		hsftp_recurs_fdownload_get_next_file(session_handle);

	return;
}

/* load local directory listing */
extern void hsftp_recurs_load_local_directory(void)
{
   	WIN32_FIND_DATA     FileData; 
   	HANDLE              hSearch;  
    BOOL                fFinished = FALSE;
    __int64				fSize;
    unsigned char		s[100];
	list_item_t			*p;

   	hSearch = FindFirstFile(TEXT("*.*"), &FileData); 
   	if (hSearch == INVALID_HANDLE_VALUE) 
   	{ 
       	return;
   	} 

   	while (!fFinished) 
   	{		
		memset(s, 0, sizeof (s));
		fSize = (__int64)((FileData.nFileSizeHigh << 32) & 0xffffffff00000000);
		fSize |= (__int64)((FileData.nFileSizeLow) & 0x00000000ffffffff);
		sprintf(s, "%I64d", fSize);
	
		p = ListPut(ctrl.pCurList, FileData.cFileName, strlen(FileData.cFileName) + 1);
		if (!p) return;

		ItemAddDataEx1(p, s, strlen(s)+1);
		ItemAddDataEx2(p, (FileData.dwFileAttributes & FILE_ATTRIBUTE_DIRECTORY) ? "d" : " ", 2);		

     	if (!FindNextFile(hSearch, &FileData)) 
      	{
         	if (GetLastError() == ERROR_NO_MORE_FILES) 
         	{ 
            	fFinished = TRUE; 
         	} 
         	else 
       		{  	 
            	return;
         	} 
      	}
   	}    

	ctrl.pCurList->pMarker = ctrl.pCurList->head;
}

/* idle state, no recursive operation in progress  */
static int hsftp_recurs_state_idle(long session_handle, int ev, long arg1, long arg2, long arg3)
{
	switch (ev)
	{
	case HSFTP_RECURS_EV_START_FOLDER_DELETE_REQ:
		hsftp_recurs_start_folder_delete(session_handle, (unsigned char *)arg1);
		break;

	case HSFTP_RECURS_EV_START_FOLDER_DOWNLOAD_REQ:
		hsftp_recurs_start_folder_download(session_handle, (unsigned char *)arg1);
		break;

	case HSFTP_RECURS_EV_START_FOLDER_UPLOAD_REQ:
		hsftp_recurs_start_folder_upload(session_handle, (unsigned char *)arg1);
		break;
	}
	
	return 0;
}

/* recursive deleting folder, waiting result of remote chdir down */
static int hsftp_recurs_state_fdel_chdir_down(long session_handle, int ev, long arg1, long arg2, long arg3)
{
	int rc;
	
	switch (ev)
	{
	case HSFTP_RECURS_EV_CWD_DONE:		
		hsftp_recurs_change_state(hsftp_recurs_state_fdel_down_waitlist);			

		rc = HsFtpCliList(session_handle);
		if (rc != HS_FTP_RC_OK)
		{
			hsftp_recurs_on_folder_delete_failed(session_handle, HSFTP_RECURS_RC_RCHDIR, arg3);
			break;
		}
		break;

	case HSFTP_RECURS_EV_CWD_FAILED:
	case HSFTP_RECURS_EV_CTRL_CLOSED:
		hsftp_recurs_on_folder_delete_failed(session_handle, HSFTP_RECURS_RC_RCHDIR, arg3);
		break;
	}
	
	return 0;
}

/* recursive downloading folder, waiting result of remote chdir down */
static int hsftp_recurs_state_fget_chdir_down(long session_handle, int ev, long arg1, long arg2, long arg3)
{
	int rc;
	
	switch (ev)
	{
	case HSFTP_RECURS_EV_CWD_DONE:
		hsftp_recurs_change_state(hsftp_recurs_state_fget_waitlist);			

		rc = HsFtpCliList(session_handle);
		if (rc != HS_FTP_RC_OK)
		{
			hsftp_recurs_on_folder_download_failed(session_handle, HSFTP_RECURS_RC_LIST, arg3);
			break;
		}			
		break;

	case HSFTP_RECURS_EV_CWD_FAILED:
	case HSFTP_RECURS_EV_CTRL_CLOSED:
		hsftp_recurs_on_folder_download_failed(session_handle, HSFTP_RECURS_RC_RCHDIR, arg3);
		break;
	}	

	return 0;
}

/* recursive upload folder: waiting result of change directory into directory we've created at remote */
static int hsftp_recurs_state_fput_chdir_down(long session_handle, int ev, long arg1, long arg2, long arg3)
{	
	switch (ev)
	{
	case HSFTP_RECURS_EV_CWD_DONE:
		hsftp_recurs_load_local_directory();
		hsftp_recurs_fupload_get_next_file(session_handle);
		break;

	case HSFTP_RECURS_EV_CWD_FAILED:
	case HSFTP_RECURS_EV_CTRL_CLOSED:
		hsftp_recurs_on_folder_upload_failed(session_handle, HSFTP_RECURS_RC_RCHDIR, arg3);
		break;
	}	

	return 0;
}

/* recursive deleting folder, waiting result of remote chdir up */
static int hsftp_recurs_state_fdel_chdir_up(long session_handle, int ev, long arg1, long arg2, long arg3)
{
	int rc;
	list_item_t *p;
	
	switch (ev)
	{
	case HSFTP_RECURS_EV_CWD_DONE:		
		// we have returned up from a folder we know is empty at this point
		// Get a pointer to list entry containing the folder name
		if (ctrl.pCurList->pMarker)
			p = (list_item_t *)ctrl.pCurList->pMarker->prev;
		else
			p = ctrl.pCurList->tail;		

		if (p)
		{
			hsftp_recurs_event_to_user(
				session_handle, HSFTP_RECURS_USR_EV_FOLDER_DELETE_STATUS, 
				(long)p->data, HSFTP_OP_START);
						
			hsftp_recurs_change_state(hsftp_recurs_state_fdel_rmdir);
			
			rc = HsFtpCliRemoveDir(session_handle, (unsigned char *)p->data);
			if (rc != HS_FTP_RC_OK)
			{
				hsftp_recurs_on_folder_delete_failed(session_handle, HSFTP_RECURS_RC_RMDIR, arg3);
				break;
			}
		}		
		break;

	case HSFTP_RECURS_EV_CWD_FAILED:
	case HSFTP_RECURS_EV_CTRL_CLOSED:
		hsftp_recurs_on_folder_delete_failed(session_handle, HSFTP_RECURS_RC_RCHDIR, arg3);
		break;
	}
	
	return 0;
}

/* recursive downloading folder: waiting result of chdir up */
static int hsftp_recurs_state_fget_chdir_up(long session_handle, int ev, long arg1, long arg2, long arg3)
{
	switch (ev)
	{
	case HSFTP_RECURS_EV_CWD_DONE:
		hsftp_recurs_fdownload_get_next_file(session_handle);
		break;

	case HSFTP_RECURS_EV_CWD_FAILED:
	case HSFTP_RECURS_EV_CTRL_CLOSED:
		hsftp_recurs_on_folder_download_failed(session_handle, HSFTP_RECURS_RC_RCHDIR, arg3);
		break;
	}	

	return 0;
}

/* recursive uploading folder: waiting result of chdir up */
static int hsftp_recurs_state_fput_chdir_up(long session_handle, int ev, long arg1, long arg2, long arg3)
{
	switch (ev)
	{
	case HSFTP_RECURS_EV_CWD_DONE:
		hsftp_recurs_fupload_get_next_file(session_handle);
		break;

	case HSFTP_RECURS_EV_CWD_FAILED:
	case HSFTP_RECURS_EV_CTRL_CLOSED:
		hsftp_recurs_on_folder_upload_failed(session_handle, HSFTP_RECURS_RC_RCHDIR, arg3);
		break;
	}	

	return 0;
}

/* recursive downloading folder, waiting result of remote chdir up to the last top level, folder download complete */
static int hsftp_recurs_state_fget_chdir_root(long session_handle, int ev, long arg1, long arg2, long arg3)
{
	switch (ev)
	{
	case HSFTP_RECURS_EV_CWD_DONE:
		ListFree(&ctrl.file_tree);

		hsftp_recurs_change_state(hsftp_recurs_state_idle);
			
		hsftp_recurs_event_to_user(
			session_handle, HSFTP_RECURS_USR_EV_FOLDER_DOWNLOAD_DONE, 0, 0);		
		break;

	case HSFTP_RECURS_EV_CWD_FAILED:
	case HSFTP_RECURS_EV_CTRL_CLOSED:
		hsftp_recurs_on_folder_download_failed(session_handle, HSFTP_RECURS_RC_RCHDIR, arg3);
		break;
	}
	
	return 0;
}

/* recursive uploading folder, waiting result of remote chdir up to the last top level, folder upload complete */
static int hsftp_recurs_state_fput_chdir_root(long session_handle, int ev, long arg1, long arg2, long arg3)
{
	switch (ev)
	{
	case HSFTP_RECURS_EV_CWD_DONE:
		ListFree(&ctrl.file_tree);

		hsftp_recurs_change_state(hsftp_recurs_state_idle);
			
		hsftp_recurs_event_to_user(
			session_handle, HSFTP_RECURS_USR_EV_FOLDER_UPLOAD_DONE, 0, 0);		
		break;

	case HSFTP_RECURS_EV_CWD_FAILED:
	case HSFTP_RECURS_EV_CTRL_CLOSED:
		hsftp_recurs_on_folder_upload_failed(session_handle, HSFTP_RECURS_RC_RCHDIR, arg3);
		break;
	}
	
	return 0;	
}

/* recursive deleting folder, waiting result of remove dir */
static int hsftp_recurs_state_fdel_rmdir(long session_handle, int ev, long arg1, long arg2, long arg3)
{
	switch (ev)
	{
	case HSFTP_RECURS_EV_RMD_DONE:		
		hsftp_recurs_event_to_user(
			session_handle, HSFTP_RECURS_USR_EV_FOLDER_DELETE_STATUS, 
			0, HSFTP_OP_DONE);
				
		// if deleted top level folder, we are all done
		if (ctrl.pCurList == &ctrl.file_tree)
		{			
			ListFree(&ctrl.file_tree);

			hsftp_recurs_change_state(hsftp_recurs_state_idle);
			
			hsftp_recurs_event_to_user(
				session_handle, HSFTP_RECURS_USR_EV_FOLDER_DELETE_DONE, 0, 0);
			break;
		}

		hsftp_recurs_fdelete_get_next_file(session_handle);
		break;

	case HSFTP_RECURS_EV_RMD_FAILED:
	case HSFTP_RECURS_EV_CTRL_CLOSED:
		hsftp_recurs_on_folder_delete_failed(session_handle, HSFTP_RECURS_RC_RMDIR, arg3);
		break;	
	}

	return 0;
}

/* recursive upload folder, waiting result of create directory on remote */
static int hsftp_recurs_state_fput_mkdir(long session_handle, int ev, long arg1, long arg2, long arg3)
{
	int rc;
	
	switch(ev)
	{
	case HSFTP_RECURS_EV_MKD_DONE:
		hsftp_recurs_change_state(hsftp_recurs_state_fput_chdir_down);

		rc = HsFtpCliChDir(session_handle, ctrl.path);
		if (rc != HS_FTP_RC_OK)
		{
			hsftp_recurs_on_folder_upload_failed(session_handle, HSFTP_RECURS_RC_RMKDIR, arg3);
		}		
		break;
		
	case HSFTP_RECURS_EV_MKD_FAILED:
	case HSFTP_RECURS_EV_CTRL_CLOSED:
		hsftp_recurs_on_folder_upload_failed(session_handle, HSFTP_RECURS_RC_RMKDIR, arg3);
		break;	
	}

	return 0;
}

/* recursive deleting folder, waiting result of individial file delete */
static int hsftp_recurs_state_fdel_file(long session_handle, int ev, long arg1, long arg2, long arg3)
{
	switch (ev)
	{
	case HSFTP_RECURS_EV_FILE_DELE_DONE:
		hsftp_recurs_event_to_user(
			session_handle, HSFTP_RECURS_USR_EV_FOLDER_DELETE_STATUS, 
			0, HSFTP_OP_DONE);
		
		hsftp_recurs_fdelete_get_next_file(session_handle);
		break;

	case HSFTP_RECURS_EV_FILE_DELE_FAILED:
	case HSFTP_RECURS_EV_CTRL_CLOSED:
		hsftp_recurs_on_folder_delete_failed(session_handle, HSFTP_RECURS_RC_FILE, arg3);
		break;	
	}

	return 0;
}

/* recursive downloading folder: waiting result of individual get file operation */
static int hsftp_recurs_state_fget_file(long session_handle, int ev, long arg1, long arg2, long arg3)
{
	int progress;
	
	switch (ev)
	{
	case HSFTP_RECURS_EV_FILE_RXDONE:	
		hsftp_recurs_event_to_user(
			session_handle, HSFTP_RECURS_USR_EV_FOLDER_PROGRESS, 100, 0);

		hsftp_recurs_event_to_user(
			session_handle, HSFTP_RECURS_USR_EV_FOLDER_DOWNLOAD_STATUS, 
			0, HSFTP_OP_DONE);

		hsftp_recurs_fdownload_get_next_file(session_handle);
		break;

	case HSFTP_RECURS_EV_FILE_RXSEG:
    	ctrl.rxed_len += arg1;
		if ((ctrl.fsize) != 0)
		{
			progress = (int)((ctrl.rxed_len * 100) / ctrl.fsize);

			hsftp_recurs_event_to_user(
				session_handle, HSFTP_RECURS_USR_EV_FOLDER_PROGRESS, (long)progress, 0);
		}		
		break;

	case HSFTP_RECURS_EV_CTRL_CLOSED:
		hsftp_recurs_on_folder_download_failed(session_handle, HSFTP_RECURS_RC_FILE, arg3);
		break;	
	}

	return 0;
}

/* recursive folder upload: waiting for completion of file upload and sending data */
static int hsftp_recurs_state_fput_file(long session_handle, int ev, long arg1, long arg2, long arg3)
{	
	switch (ev)
	{
	case HSFTP_RECURS_EV_FILE_TXDONE:	
		hsftp_recurs_event_to_user(
			session_handle, HSFTP_RECURS_USR_EV_FOLDER_PROGRESS, 100, 0);

		hsftp_recurs_event_to_user(
			session_handle, HSFTP_RECURS_USR_EV_FOLDER_UPLOAD_STATUS, 
			0, HSFTP_OP_DONE);

		hsftp_recurs_fupload_get_next_file(session_handle);
		break;

	case HSFTP_RECURS_EV_FILE_TXSEG:
	    ctrl.txed_len += arg1;
		if (ctrl.fsize != 0)
		{
			int progress;

			progress = (int)((ctrl.txed_len * 100) / ctrl.fsize);
			hsftp_recurs_event_to_user(
				session_handle, HSFTP_RECURS_USR_EV_FOLDER_PROGRESS, (long)progress, 0);		
		}		
		break;

	case HSFTP_RECURS_EV_CTRL_CLOSED:
		hsftp_recurs_on_folder_upload_failed(session_handle, HSFTP_RECURS_RC_FILE, arg3);
		break;	
	}

	return 0;
}

/* deleting folder wait for listing after having changed into remote directory  */
static int hsftp_recurs_state_fdel_down_waitlist(long session_handle, int ev, long arg1, long arg2, long arg3)
{
	switch (ev)
	{
	case HSFTP_RECURS_EV_LIST:		
		hsftp_recurs_fdel_process_listing_chdir_down(session_handle, (unsigned char *)arg1, (int)arg2);
		break;

	case HSFTP_RECURS_EV_CTRL_CLOSED:
		hsftp_recurs_on_folder_delete_failed(session_handle, HSFTP_RECURS_RC_LIST, arg3);
		break;		
	}
	
	return 0;
}

/* dwonloading folder: wait for listing after having changed into remote directory  */
static int hsftp_recurs_state_fget_waitlist(long session_handle, int ev, long arg1, long arg2, long arg3)
{
	switch (ev)
	{
	case HSFTP_RECURS_EV_LIST:
		hsftp_recurs_fdel_process_listing_chdir_down(session_handle, (unsigned char *)arg1, (int)arg2);
		break;

	case HSFTP_RECURS_EV_CTRL_CLOSED:
		hsftp_recurs_on_folder_download_failed(session_handle, HSFTP_RECURS_RC_LIST, arg3);
		break;		
	}	

	return 0;
}

/* feed local events into recursive state machine */
static void hsftp_recurs_do_fsm_event(long session_handle, int ev, long arg1, long arg2, long arg3)
{
	ctrl.fsm(session_handle, ev, arg1, arg2, arg3);
}

/* Initialise recursive operations module */
extern void HsFtpRecursInit(hsftp_recursive_callback_t *cb)
{
	ctrl.fsm = hsftp_recurs_state_idle;
	ctrl.cb = cb;
	ctrl.initialized = TRUE;
}

/* Release resources used by recursive folder operations module */
extern void HsFtpRecurseCleanUp(void)
{
	long session_handle;
	int ev, core_ev;
	unsigned char *pBuf = NULL;
	int len = 0;
	
	if (!ctrl.initialized)
		return;

	ListFree(&ctrl.file_tree);

	while (1)
	{
		pBuf = NULL;
		len = 0;
		
		if (hsftp_recurs_qget(&ctrl.evtq, &session_handle, &ev, &core_ev, &pBuf, &len))
		{			
			if ((ev == HSFTP_RECURS_EV_LIST) && (pBuf))
				free(pBuf);
		}
		else
		{
			break;
		}
		
	}

	ctrl.initialized = FALSE;
}

/*
 * Feed events from core FTP library 
 */
extern void HsFtpRecursDoEvent(
	long 		session_handle,		// FTP session handle
	int 		ev, 				// event
	long 		arg1, 				// argument 1
	long 		arg2, 				// argument 2
	long 		arg3)				// argument 3
{
	unsigned char *pBuf = NULL;
	int len = 0;
	int local_ev;
	
	if (!ctrl.initialized)
		return;

	if (!ctrl.fsm)
		return;

	switch (ev)
	{
   	case HS_FTPCLI_USR_EV_CLOSED:        // FTP control connection closed
   	case HS_FTPCLI_USR_EV_SRVERR:        // FTP session closed due to server error reply
   	case HS_FTPCLI_USR_EV_UNEXP:         // Unexpected server response, session closed
   	case HS_FTPCLI_USR_EV_UNKNOWN:		 // Unrecognized server response
   	case HS_FTPCLI_USR_EV_NOCTX:         // FTP session closed, no free contexts
   	case HS_FTPCLI_USR_EV_TIMEDOUT:      // FTP session closed due to timeout
   	case HS_FTPCLI_USR_EV_NOMEM:         // FTP session closed, no free memory
   	case HS_FTPCLI_USR_EV_RXINCOMPL:     // File receive failed, incomplete.
   	case HS_FTPCLI_USR_EV_TYP_FAILED:	 // setting transfer type failed
   	case HS_FTPCLI_USR_EV_ABORTED:		 // file transfer aborted, can be for send or receive transfer
	case HS_FTPCLI_USR_EV_STOR_FAILED:
   		local_ev = HSFTP_RECURS_EV_CTRL_CLOSED;
   		break;
		    	
   	case HS_FTPCLI_USR_EV_LIST:          // current directory listing
   		pBuf = (unsigned char *)arg1;
		len = (int)arg2;
   		local_ev = HSFTP_RECURS_EV_LIST;		
   		break;
    	
   	case HS_FTPCLI_USR_EV_CWD_FAILED:	 // HsFtpCliChDir function (CWD) request failed
   		local_ev = HSFTP_RECURS_EV_CWD_FAILED;
		break;
		
   	case HS_FTPCLI_USR_EV_CWD_DONE:      // HsFtpCliChDir function (CWD) complete
   		local_ev = HSFTP_RECURS_EV_CWD_DONE;
		break;

   	case HS_FTPCLI_USR_EV_RX_STATUS:         // receive segment of file
   		hsftp_recurs_do_fsm_event(session_handle, HSFTP_RECURS_EV_FILE_RXSEG, arg3, arg2, (long)ev);
		return;

   	case HS_FTPCLI_USR_EV_TX_STATUS:		// transmit segment of file
   		hsftp_recurs_do_fsm_event(session_handle, HSFTP_RECURS_EV_FILE_TXSEG, arg3, arg2, (long)ev);
		return;
		
   	case HS_FTPCLI_USR_EV_RXDONE:        // File receive transfer complete
   		local_ev = HSFTP_RECURS_EV_FILE_RXDONE;
		break;
		
   	case HS_FTPCLI_USR_EV_TXDONE:		 // File send transfer complete
   		local_ev = HSFTP_RECURS_EV_FILE_TXDONE;
		break;

	case HS_FTPCLI_USR_EV_MKD_FAILED:	// HsFtpCliCreateDir function (MKD) request failed		
		local_ev = HSFTP_RECURS_EV_MKD_FAILED;
		break;
	
   	case HS_FTPCLI_USR_EV_MKD_DONE:     // HsFtpCliCreateDir function (MKD) complete
		local_ev = HSFTP_RECURS_EV_MKD_DONE;
		break;
		
   	case HS_FTPCLI_USR_EV_RMD_FAILED:	// HsFtpCliRemoveDir function (RMD) request failed    
   		local_ev = HSFTP_RECURS_EV_RMD_FAILED; 
		break;
		
   	case HS_FTPCLI_USR_EV_RMD_DONE:     // HsFtpCliRemoveDir function (RMD) complete
   		local_ev = HSFTP_RECURS_EV_RMD_DONE;
		break;
		
   	case HS_FTPCLI_USR_EV_DELE_FAILED:	// HsFtpCliDeleteFile function (DELE) request failed    
   		local_ev = HSFTP_RECURS_EV_FILE_DELE_FAILED;
		break;
		
   	case HS_FTPCLI_USR_EV_DELE_DONE:    // HsFtpCliDeleteFile function (DELE) complete
   		local_ev = HSFTP_RECURS_EV_FILE_DELE_DONE;
   		break;

	default:
		return;
	}

	hsftp_recurs_qput(&ctrl.evtq, session_handle, local_ev, ev, pBuf, len);
}

/*
 * Tick function
 */
extern void HsFtpRecursTick(void)
{	
	long session_handle;
	int ev, core_ev;
	unsigned char *pBuf = NULL;
	int len = 0;
	
	while (1)
	{
		pBuf = NULL;
		len = 0;
		
		if (hsftp_recurs_qget(&ctrl.evtq, &session_handle, &ev, &core_ev, &pBuf, &len))
		{
			hsftp_recurs_do_fsm_event(session_handle, ev, (long)pBuf, (long)len, (long)core_ev);
			
			if ((ev == HSFTP_RECURS_EV_LIST) && (pBuf))
				free(pBuf);
		}
		else
		{
			break;
		}
	}
}

/*
 * Delete recursively complete folder with all subfolders and files
 */
extern int HsFtpRecursDeleteFolder(
	long 		session_handle, 	// FTP session handle
	unsigned char *foldername)		// folder to delete
{
	int length;
	
	if (!ctrl.initialized)
		return HSFTP_RECURS_RC_NOTINIT;

	if ((!foldername) || (!session_handle))
		return HSFTP_RECURS_RC_INVPAR;

	length = (int)strlen(foldername);
	if (!length) return HSFTP_RECURS_RC_INVPAR;

	if (ctrl.fsm != hsftp_recurs_state_idle)
		return HSFTP_RECURS_RC_BUSY;

	hsftp_recurs_do_fsm_event(session_handle, 
		HSFTP_RECURS_EV_START_FOLDER_DELETE_REQ, (long)foldername, 0, 0);

	return HSFTP_RECURS_RC_OK;
}

/*
 * Download recursively complete folder with all subfolders and files
 */
extern int HsFtpRecursDownloadFolder(
	long 		session_handle, 	// FTP session handle
	unsigned char *foldername,		// folder to download
	int			overwrite)			// if true, overwrite local files and folders
									// if false, overwrite only if local and 
									// remote files size don't match
{
	int length;
	
	if (!ctrl.initialized)
		return HSFTP_RECURS_RC_NOTINIT;

	if ((!foldername) || (!session_handle))
		return HSFTP_RECURS_RC_INVPAR;

	length = (int)strlen(foldername);
	if (!length) return HSFTP_RECURS_RC_INVPAR;

	if (ctrl.fsm != hsftp_recurs_state_idle)
		return HSFTP_RECURS_RC_BUSY;

	getcwd(ctrl.start_dir, sizeof(ctrl.start_dir));

	ctrl.overwrite = overwrite;

	hsftp_recurs_do_fsm_event(session_handle, 
		HSFTP_RECURS_EV_START_FOLDER_DOWNLOAD_REQ, (long)foldername, 0, 0);

	return HSFTP_RECURS_RC_OK;
}

/*
 * Upload recursively complete folder with all subfolders and files
 */
extern int HsFtpRecursUploadFolder(
	long 		session_handle, 	// FTP session handle
	unsigned char *foldername)		// folder to upload
{
	int length;
	
	if (!ctrl.initialized)
		return HSFTP_RECURS_RC_NOTINIT;

	if ((!foldername) || (!session_handle))
		return HSFTP_RECURS_RC_INVPAR;

	length = (int)strlen(foldername);
	if (!length) return HSFTP_RECURS_RC_INVPAR;

	if (length >= (sizeof(ctrl.path)))
		return HSFTP_RECURS_RC_INVPAR;

	memset(ctrl.path, 0, sizeof(ctrl.path));
	memcpy(ctrl.path, foldername, length);

	if (ctrl.fsm != hsftp_recurs_state_idle)
		return HSFTP_RECURS_RC_BUSY;

	getcwd(ctrl.start_dir, sizeof(ctrl.start_dir));

	if (chdir(foldername) != 0)
		return HSFTP_RECURS_RC_LCHDIR;

	hsftp_recurs_do_fsm_event(session_handle, 
		HSFTP_RECURS_EV_START_FOLDER_UPLOAD_REQ, (long)foldername, 0, 0);

	return HSFTP_RECURS_RC_OK;
}


