/*
 * HS FTP Client Library
 * Recursive folder related operations
 * Developed by Hillstone Software
 * http://www.hillstone-software.com
 * support@hillstone-software.com
 */
#ifndef _HSFTP_RECURS_H
#define _HSFTP_RECURS_H

#include <malloc.h>
#include <string.h>
#include <ctype.h>
#include <stdlib.h>
#include <stdio.h>
#include <direct.h>
#include <windows.h>
#include "HsList.h"
#include "HsFtp_if.h"
#include "HsFtpRecurs_if.h"

#ifndef FALSE
#define FALSE 0
#define TRUE 1
#endif

/* recursive FSM events */
typedef enum
{
	HSFTP_RECURS_EV_START_FOLDER_DELETE_REQ,				// initiate folder delete
	HSFTP_RECURS_EV_START_FOLDER_DOWNLOAD_REQ,				// initiate folder download
	HSFTP_RECURS_EV_START_FOLDER_UPLOAD_REQ,				// initiate folder upload
	HSFTP_RECURS_EV_CWD_DONE,								// remote chdir complete
	HSFTP_RECURS_EV_CWD_FAILED,								// remote chdir failed
	HSFTP_RECURS_EV_FILE_DELE_DONE,							// file delete complete
	HSFTP_RECURS_EV_FILE_DELE_FAILED,						// file delete complete
	HSFTP_RECURS_EV_LIST,									// remote listing received
	HSFTP_RECURS_EV_RMD_DONE,								// directory removed at remote
	HSFTP_RECURS_EV_MKD_DONE,								// directory created at remote
	HSFTP_RECURS_EV_RMD_FAILED,								// directory remove failed
	HSFTP_RECURS_EV_MKD_FAILED,								// directory create failed
	HSFTP_RECURS_EV_CTRL_CLOSED,							// control connection closed
	HSFTP_RECURS_EV_FILE_RXSEG,								// received next file segment
	HSFTP_RECURS_EV_FILE_TXSEG,								// request for next file segment to transmit
	HSFTP_RECURS_EV_FILE_TXDONE,							// file transmission complete
	HSFTP_RECURS_EV_FILE_RXDONE,							// file reception complete
	HSFTP_RECURS_EV_ABORT,									// operation aborted by user
} hsftp_recurs_fsm_ev_t;

typedef int hs_ftp_recurs_fsm_t(long session_handle, int ev, long arg1, long arg2, long arg3);

typedef struct
{
	int							ev;
	int							core_ev;
	long						session_handle;
	unsigned char				*pBuf;
	int							len;
} recurs_evt_q_cell_t;

#define HSFTP_RECURS_MAXQ		10
typedef struct
{
	int							in;
	int							out;
	recurs_evt_q_cell_t			cells[HSFTP_RECURS_MAXQ];
} recurs_q_t;

typedef struct
{
	hs_ftp_recurs_fsm_t			*fsm;			// recursive state machine
	hsftp_recursive_callback_t 	*cb;			// user callback
	int							initialized;	// initialised flag
	list_t						file_tree;		// file tree structure
	list_t						*pCurList;		// current level pointer
	recurs_q_t					evtq;			// event queue, receives events from core HSFTP
	unsigned char				start_dir[_MAX_PATH]; // starting directory when dowload operation initiated
	__int64						rxed_len;		// received so far
	__int64						txed_len;		// sent so far
	__int64						fsize;			// total file size
	unsigned char				path[_MAX_PATH];// filename fo upload
	int							overwrite;		// 1=overwrite files; 0=overwrite only if local and remote file size mismatch
} hsftprecurs_t;

static int hsftp_recurs_state_idle(long session_handle, int ev, long arg1, long arg2, long arg3);
static int hsftp_recurs_state_fdel_chdir_down(long session_handle, int ev, long arg1, long arg2, long arg3);
static int hsftp_recurs_state_fdel_chdir_up(long session_handle, int ev, long arg1, long arg2, long arg3);
static int hsftp_recurs_state_fdel_rmdir(long session_handle, int ev, long arg1, long arg2, long arg3);
static int hsftp_recurs_state_fdel_file(long session_handle, int ev, long arg1, long arg2, long arg3);
static int hsftp_recurs_state_fdel_down_waitlist(long session_handle, int ev, long arg1, long arg2, long arg3);
static int hsftp_recurs_state_fget_chdir_down(long session_handle, int ev, long arg1, long arg2, long arg3);
static int hsftp_recurs_state_fget_chdir_root(long session_handle, int ev, long arg1, long arg2, long arg3);
static int hsftp_recurs_state_fget_chdir_up(long session_handle, int ev, long arg1, long arg2, long arg3);
static int hsftp_recurs_state_fget_file(long session_handle, int ev, long arg1, long arg2, long arg3);
static int hsftp_recurs_state_fget_waitlist(long session_handle, int ev, long arg1, long arg2, long arg3);
static int hsftp_recurs_state_fput_mkdir(long session_handle, int ev, long arg1, long arg2, long arg3);
static int hsftp_recurs_state_fput_chdir_root(long session_handle, int ev, long arg1, long arg2, long arg3);
static int hsftp_recurs_state_fput_chdir_up(long session_handle, int ev, long arg1, long arg2, long arg3);
static int hsftp_recurs_state_fput_file(long session_handle, int ev, long arg1, long arg2, long arg3);

#endif /* _HSFTP_RECURS_H */
