/*
 * Developed by Hillstone Software. http://www.hillstone-software.com
 * HS NTP Protocol Module Demo application
 */
#include "stdafx.h"
#include <commctrl.h>
#include <Commdlg.h>
#include "resource.h"
#include <stdlib.h>
#include <direct.h>
#include <stdio.h>
#include <process.h>
#include <time.h>
#include <Windowsx.h>
#include <Mmsystem.h>

#include "hsip_if.h"
#include "hsicmp_if.h"
#include "hssock_if.h"
#include "hsdns_if.h"
#include "hsdhcp_if.h"
#include "pcap.h"

#define MAX_EVENTS 1000

#define WM_APP_TIMER	WM_USER + 1		// timer messages

#define MAXTIMERS		256

typedef struct
{
	int					allocated;
	int					enabled;
	void 				*arg;
	hs_sock_timer_cb_t 	*cb;
	UINT				timer_id;
} timerbock_t;

#define HS_VERSION	"v1.1"

TCHAR			szWindowClass[]		= " ";
HINSTANCE		hInst;
TCHAR			szTitle[] = "HsTCPIPv4 C Source Library: ICMP Ping Demo";

HWND			hwnd_main;
HWND			dlg_main;
HBRUSH			hbrush1				= NULL;

static long dns_timer = 0;
static unsigned char dest_host[256] = {0};
static unsigned long timeout_ms;
static int num_pings;
static unsigned short ping_size = 32;

char commercial_msg[] = "This is a demo application for HsTCPIPv4 embedded protocol stack\n"
                        "by Hillstone Software. HS ICMP is part of HsTCPIPv4. Further information:\n"
                        "www.hillstone-software.com/hs_tcpipv4_details.htm\nemail: info@hillstone-software.com; Tel. +353 87 988 1568";

// pcap local interface device list
pcap_if_t *alldevs = NULL;
pcap_t *adhandle = NULL;

static 	HFONT			hfont				= NULL;
static 	LOGFONT			fnt;

unsigned char server_name[256] = "www.google.com";

unsigned char port_str[5] = {0};
unsigned char regdata[1024] = {0};

conf_t cfg = 
{
	{DEF_PORT},				// port
	{0xC0A80101},			// dest ip address
	{"ntp.server.com"},
};

static COLORREF color;

static timerbock_t	timers[MAXTIMERS] = {0};

static TIMECAPS		timecaps;

// allocate timer
static timerbock_t *alloc_timer(void)
{
	int i;

	for (i=0; i<MAXTIMERS; i++)
	{
		if (!timers[i].allocated)
		{
			timers[i].allocated = TRUE;
			return &timers[i];
		}
	}

	return NULL;
}

// free timer
static void free_timer(timerbock_t *p)
{
	p->allocated = FALSE;
}

HWND NewDialogWin(HINSTANCE hInst, HWND parent, DLGPROC proc, int dlg_id)
{
	HWND	handle;

	handle = CreateDialog(hInst, (LPCTSTR)dlg_id, parent, proc);

	ShowWindow	(handle, SW_SHOWDEFAULT);
    UpdateWindow(handle);

	return	handle;
}

ATOM BlRegisterClass(HINSTANCE	hInstance,
					UINT		style,
					WNDPROC		proc,
					int			cursor_id,
					HBRUSH		bkg_brush,
					LPCTSTR		lpszClassName,
					int			icon_id,
					int			menu_id)
{
	WNDCLASSEX	wcex;

	wcex.cbSize			= sizeof(WNDCLASSEX); 

	wcex.style			= style;
	wcex.lpfnWndProc	= proc;
	wcex.cbClsExtra		= 0;
	wcex.cbWndExtra		= 0;
	wcex.hInstance		= hInstance;
	wcex.hIcon			= 0;

	wcex.hCursor		= 0;
	if (cursor_id)
	{
		wcex.hCursor	= LoadCursor(0, (const char *)cursor_id);
	}

	wcex.hbrBackground	= bkg_brush;
	wcex.lpszMenuName	= (LPCSTR)menu_id;
	wcex.lpszClassName	= lpszClassName;

	wcex.hIconSm		= 0; 
	if (icon_id)
	{
		wcex.hIcon		= LoadIcon(hInstance, (LPCTSTR)icon_id);
	}
	return RegisterClassEx(&wcex);
}

/*
 * Retrieve configuration from registry, if not found create
 */
void get_config(void)
{
	LONG	res;
	HKEY	hKey = NULL;

	HKEY hk; 
   	DWORD dwDisp, size;

	/* Check if our registry key exists */
	res = RegOpenKeyEx(HKEY_LOCAL_MACHINE, 
					"Software\\Hillstone Software\\HsPing\\1.1",
            		0, KEY_QUERY_VALUE, &hKey );

    if (res == ERROR_SUCCESS)
    {    
		RegQueryValueEx(hKey, "RegData", 0, &dwDisp, (LPBYTE) regdata, &size);

   		memcpy(&cfg, regdata, sizeof(conf_t));
   		
		memset(server_name, 0, sizeof server_name);
		memcpy(server_name, cfg.srvname, sizeof cfg.srvname);

    	RegCloseKey(hKey);
	}
	else if (RegCreateKeyEx(HKEY_LOCAL_MACHINE, "Software\\Hillstone Software\\HsPing\\1.1", 
          	0, NULL, REG_OPTION_NON_VOLATILE, KEY_WRITE, NULL, &hk, &dwDisp)) 
   	{
   		return;
   	}
   	else
   	{	
   		memcpy(regdata, &cfg, sizeof(conf_t));
		RegSetValueEx(hk, "RegData", 0, REG_BINARY, (LPBYTE) regdata, sizeof(conf_t));		
    	RegCloseKey(hk);
	}
}

/*
 * Update configuration in registry
 */
void update_config(void)
{
	LONG	res;
	HKEY	hKey = NULL;

	HKEY hk; 
   	DWORD dwDisp;
	BOOL rs = FALSE;

	SendMessage(GetDlgItem(dlg_main, IDC_IPADDRESS2), IPM_GETADDRESS, 0,(LPARAM)&cfg.dest_ip);

	cfg.port = (WORD)GetDlgItemInt(dlg_main, IDC_EDIT_PORT, &rs, FALSE);

	memset(cfg.srvname, 0, sizeof cfg.srvname);
	GetDlgItemText(dlg_main, IDC_EDIT_SRVNAME, cfg.srvname, (sizeof cfg.srvname) - 1);	

	memset(cfg.username, 0, sizeof cfg.username);
    GetDlgItemText(dlg_main, IDC_EDIT_USERNAME, cfg.username, (sizeof cfg.username) - 1);

	memset(cfg.password, 0, sizeof cfg.password);
    GetDlgItemText(dlg_main, IDC_EDIT_PASSWORD, cfg.password, (sizeof cfg.password) - 1);
	

	/* Check if our registry key exists */
	res = RegOpenKeyEx(HKEY_LOCAL_MACHINE, 
					"Software\\Hillstone Software\\HsPing\\1.1",
            		0, KEY_ALL_ACCESS, &hKey );

    if (res == ERROR_SUCCESS)
    {    
   		memcpy(regdata, &cfg, sizeof(conf_t));
   		RegSetValueEx(hKey, "RegData", 0, REG_BINARY, (LPBYTE) regdata, sizeof(conf_t));		
    	RegCloseKey(hKey);
	}
	else if (RegCreateKeyEx(HKEY_LOCAL_MACHINE, "Software\\Hillstone Software\\HsPing\\1.1", 
          	0, NULL, REG_OPTION_NON_VOLATILE, KEY_WRITE, NULL, &hk, &dwDisp)) 
   	{
   		return;
   	}
   	else
   	{
   		memcpy(regdata, &cfg, sizeof(conf_t));
		RegSetValueEx(hk, "RegData", 0, REG_BINARY, (LPBYTE) regdata, sizeof(conf_t));		
    	RegCloseKey(hk);
	}
}

static void LvInsertCol(HWND hwndListView, int index, unsigned char *hdr, int cx, int fmt)
{
    LV_COLUMN lvColumn = {0};
    
    lvColumn.mask = LVCF_FMT | LVCF_WIDTH | LVCF_TEXT | LVCF_SUBITEM;
   	lvColumn.pszText = hdr;
	lvColumn.cx = cx;
	lvColumn.fmt = fmt;
    ListView_InsertColumn (hwndListView, index, &lvColumn);    
}

static void InitLVlog(HWND hwndListView)
{
  	// Empty the list in list view.
  	ListView_DeleteAllItems (hwndListView);

  	// Insert the five columns in the list view.
    LvInsertCol(hwndListView, 0, "Operation Log", 512, LVCFMT_LEFT);

	ListView_SetExtendedListViewStyle(hwndListView, LVS_EX_FULLROWSELECT);
}

void write_event(char *s)
{
    static int num_events = 0;
    
    HWND lvhan = GetDlgItem(dlg_main, IDC_LIST_LOG);
	LVITEM item = {0};
	int index;

	memset(&item, 0, sizeof (LVITEM));
		
	item.mask = LVIF_TEXT;
	
	item.pszText = s;
	item.iItem = 0xffff;

    index = ListView_InsertItem(lvhan, (LPLVITEM)&item);

	ListView_EnsureVisible(lvhan, index, 0);;

	if (num_events < MAX_EVENTS)
	{
		num_events++;
	}
	else
	{
		ListView_DeleteItem(lvhan, 0);
	}
    
}

static unsigned char hex2byte(unsigned char *buf)
{
	unsigned char hex[] = "0123456789ABCDEF";
	unsigned char b = 0;
    unsigned char b1, b2;
	int i;

    b1 = toupper(buf[0]);
    b2 = toupper(buf[1]);

	for (i = 0; i<16; i++)
	{
		if (b1 == hex[i])
		{
			b &= 0x0f;
			b = (unsigned char)((i << 4) & 0xf0);
			break;
		}
	}

	for (i = 0; i<16; i++)
	{
		if (b2 == hex[i])
		{
			b &= 0xf0;
			b |= (unsigned char)(i & 0x0f);
			break;
		}
	}
	
	return b;
}

// Convert ascii hex ethernet address in the form 00:00:00:00:00:00 to binary 6 byte buffer
static void convert_hex_ethaddr_to_binary(unsigned char *hexbuf, unsigned char *binbuf)
{
	int i, n, maxlen, done;

	maxlen = 17;

	if (strlen(hexbuf) != 17)
		return;

	done = FALSE;
	i = 0; n = 0;
	while (!done)
	{
		binbuf[i] = hex2byte(&hexbuf[n]);

		i++;
		n += 3;

		if (i == 6) 
			done = TRUE;
	}
}

// multimedia timer callback
void CALLBACK mmTimeProc(UINT uID, UINT uMsg, DWORD dwUser, DWORD dw1, DWORD dw2)
{
	PostMessage(hwnd_main, WM_APP_TIMER, 0, (LPARAM)dwUser);
}

// Process timer message
static void process_timer(LPARAM lParam)
{
	timerbock_t *pTmrBlk;

	pTmrBlk = (timerbock_t *)lParam;

	if (!pTmrBlk) return;

	if (!pTmrBlk->allocated)
		return;

	if (!pTmrBlk->enabled)
		return;

	pTmrBlk->cb(pTmrBlk->arg);

	free_timer(pTmrBlk);
}

// start timer function
static long sock_timer_start(unsigned long timeout_ms, void *arg, hs_sock_timer_cb_t *cb)
{
	MMRESULT res;
	timerbock_t *pTmrBlk;

	pTmrBlk = (timerbock_t *)alloc_timer();
	if (!pTmrBlk) return 0;

	pTmrBlk->arg = arg;
	pTmrBlk->cb = cb;
	pTmrBlk->enabled = TRUE;

	res = timeSetEvent((UINT)timeout_ms, timecaps.wPeriodMin, mmTimeProc, (DWORD)pTmrBlk, TIME_ONESHOT);
	if (res == 0) 
	{	
		free_timer(pTmrBlk);
		return 0;
	}

	pTmrBlk->timer_id = (UINT)res;

	return (long)pTmrBlk;
}

// stop timer function
static void sock_timer_stop(long timer_han)
{
	timerbock_t *pTmrBlk;

	pTmrBlk = (timerbock_t *) timer_han;

	if (!pTmrBlk) return;

	if (!pTmrBlk->enabled) return;

	timeKillEvent(pTmrBlk->timer_id);
	pTmrBlk->enabled = FALSE;
}

// transmit over physical layer (ethernet)
static void sock_drv_tx_fn(unsigned char *pkt, int len)
{
	if (!adhandle) return;

	/* Send down the packet */
	if (pcap_sendpacket(adhandle,	// Adapter
		pkt,						// buffer with the packet
		len							// size
		) != 0)
	{
		return;
	}
}

// callback used by HSTCP to get number of millisecond ticks since bootup
// Used for generating pseudo random initial sequence number
// if callback pointer not initialized, HSTCP will use default 
// pseudo random number generator to produce initial sequence numbers
static unsigned long hs_sock_get_ticks(void)
{
	return (unsigned long)GetTickCount();
}

// Callback used by the stack whenever it needs a random 32 bit number
static unsigned long hs_sock_get_rand32(void)
{
	return (unsigned long)rand();
}

// global socket events
static long hs_sock_global_event(long user_handle, int ev, long arg1, long arg2, long arg3, long arg4)
{
	hs_dhcp_update_t *p;
	char str[17];

	switch (ev)
	{
	case HS_SOCK_EV_DHCP_UPDATE:
		p = (hs_dhcp_update_t *)arg1;
		if (!p) return 0;

		memset(str, 0, sizeof(str));
		HsSockInetNtoa(p->local_ip, str);
		SetDlgItemText(dlg_main, IDC_IPADDRESS1, str);

		memset(str, 0, sizeof(str));
		HsSockInetNtoa(p->dns_ip, str);
		SetDlgItemText(dlg_main, IDC_IPADDRESS_DNS, str);		

		memset(str, 0, sizeof(str));
		HsSockInetNtoa(p->router_ip, str);
		SetDlgItemText(dlg_main, IDC_IPADDRESS_GW, str);

		memset(str, 0, sizeof(str));
		HsSockInetNtoa(p->subnet_mask, str);
		SetDlgItemText(dlg_main, IDC_IPADDRESS_MASK, str);		
		break;
	}

	return 0;
}

/*
 * ICMP callback
 */
static long icmp_event_callback(int ev, long arg1, long arg2)
{
	unsigned char str[256];

	switch (ev)
	{
	case HS_ICMP_EV_PING_TIMEOUT:
		sprintf(str, "Ping %u timed out", arg1);
		write_event(str);
		break;
		
	case HS_ICMP_EV_PING_SENT:
		sprintf(str, "Sending Ping %u ...", arg1);
		write_event(str);
		break;
		
	case HS_ICMP_EV_PING_REPLY:
		sprintf(str, "Received Reply %u .", arg1);
		write_event(str);
		break;
		
	case HS_ICMP_EV_PING_FINISHED:
		sprintf(str, "Ping finished. Requests sent %u. Replies received %u .", arg1, arg2);
		write_event(str);
		break;
	}

	return 0;
}

// time wait finished for DNS to resolve destination host IP address, start ping now
static void dns_timeout(void *arg)
{
	int dns_rc, ping_rc;
	unsigned long addr;
	unsigned char ipstr[17] = {0};
	unsigned char str[512] = {0};

	if (dns_timer == 0)
		return;

	dns_timer = 0;

	dns_rc = HsDnsGetIpbyName(dest_host, &addr);
	if (dns_rc != HS_DNS_RC_OK)
	{
		write_event("Ping failed: Cannot resolve host name to IP address");
		return;
	}

	ping_rc = HsIcmpPing(addr, timeout_ms, num_pings, ping_size);
	if (ping_rc == HS_ICMP_RC_OK)
	{
		HsSockInetNtoa(addr, ipstr);
	
		sprintf(str, "Pinging destination %s ... (%s)", dest_host, ipstr);			
		write_event(str);
		return;
	}

	sprintf(str, "failed to start ICMP ping, rc=%u ", ping_rc);
	write_event(str);
}

/*
 * Start pinging destination
 */
static void start_ping(HWND hDlg)
{
	int rc, dns_rc, ping_rc;

	unsigned char str[512] = {0};
	unsigned long addr;
	unsigned char ipstr[17] = {0};
	BOOL bTranslated;

	GetDlgItemText(hDlg, IDC_EDIT_DEST, dest_host, sizeof(dest_host));

	timeout_ms = (unsigned long)GetDlgItemInt(hDlg, IDC_EDIT_TM, &bTranslated, FALSE);
	if (!bTranslated) return;

	num_pings = (int)GetDlgItemInt(hDlg, IDC_EDIT_NPINGS, &bTranslated, FALSE);
	if (!bTranslated) return;

	ping_size= (int)GetDlgItemInt(hDlg, IDC_EDIT_PS, &bTranslated, FALSE);
	if (!bTranslated) return;

	rc = HsSockInetAddr(dest_host, &addr);

	// not an ip address, try to resolve it using DNS module
	if (!rc)
		dns_rc = HsDnsGetIpbyName(dest_host, &addr);

	if ((rc) || ((!rc) && (dns_rc == HS_DNS_RC_OK)))
	{
		ping_rc = HsIcmpPing(addr, timeout_ms, num_pings, ping_size);
		if (ping_rc == HS_ICMP_RC_OK)
		{
			HsSockInetNtoa(addr, ipstr);
		
			sprintf(str, "Pinging destination %s ... (%s)", dest_host, ipstr);			
			write_event(str);
		}
		else if (ping_rc == HS_ICMP_RC_BUSY)
		{
			write_event("Ping already in progress");			
		}
		else
		{
			sprintf(str, "failed to start ICMP ping, rc=%u ", ping_rc);
			write_event(str);
		}
	}
	else if ((!rc) && (dns_rc == HS_DNS_RC_PENDING))
	{
		sprintf(str, "Resolving ipaddr of %s ... ", dest_host);
		write_event(str);

		dns_timer = sock_timer_start (1000, 0, dns_timeout);
	}
	else
	{
		sprintf(str, "DNS error %d ... ", dns_rc);
		write_event(str);
	}
}

// Stop ping
static void stop_ping(HWND hDlg)
{
	HsIcmpCancelPing();	
}

/*
 * Main dialog handler
 */
LRESULT CALLBACK DlgWin(HWND hDlg, UINT message, 
						  WPARAM wParam, LPARAM lParam)
{
	LRESULT result = FALSE;
	int rc, inum, i;

	unsigned char ascii_hex_eth[19] = {0};
	unsigned char ethaddr[6];

	unsigned char source_ipaddr[17] = {0};
	unsigned char gw_ipaddr[17] = {0};
	unsigned char dns_ipaddr[17] = {0};
	unsigned char mask[17] = {0};

	char str[256];

	pcap_if_t *d;

	char errbuf[PCAP_ERRBUF_SIZE];

	hssock_init_t sockinit = {0};

	switch (message)
	{
	case WM_INITDIALOG:   
        SendDlgItemMessage(hDlg, IDC_EDIT_SRVNAME, EM_LIMITTEXT, MAX_SRVNAME, 0);
        SendDlgItemMessage(hDlg, IDC_EDIT_PORT, EM_LIMITTEXT, 4, 0);
        
        SetDlgItemText(hDlg, IDC_EDIT_SRVNAME, server_name);        
        
		SendDlgItemMessage(hDlg, IDC_STATIC_CM, WM_SETTEXT, 0, (long)commercial_msg);

		dlg_main = hDlg;
		
		SendDlgItemMessage(hDlg, IDC_EDIT_RESP, WM_SETFONT, (WPARAM)hfont, TRUE);

		SetDlgItemText(hDlg, IDC_IPADDRESS1, "192.168.1.51");
		SetDlgItemText(hDlg, IDC_EDIT_MAC, "00:1e:68:a4:fb:15");

		SetDlgItemText(hDlg, IDC_IPADDRESS_GW, "192.168.1.254");
		SetDlgItemText(hDlg, IDC_IPADDRESS_MASK, "255.255.255.0");

		SetDlgItemText(hDlg, IDC_IPADDRESS_DNS, "0.0.0.0");

		SetDlgItemText(hDlg, IDC_EDIT_DEST, "www.google.com");
		SetDlgItemText(hDlg, IDC_EDIT_NPINGS, "5");
		SetDlgItemText(hDlg, IDC_EDIT_TM, "3000");
		SetDlgItemText(hDlg, IDC_EDIT_PS, "32");

		CheckDlgButton(hDlg, IDC_CHECK_DHCP, BST_CHECKED);
		EnableWindow(GetDlgItem(hDlg, IDC_IPADDRESS1), FALSE);

		InitLVlog(GetDlgItem(hDlg, IDC_LIST_LOG));

		/* Retrieve the device list */
		if(pcap_findalldevs(&alldevs, errbuf) == -1)
		{
			sprintf(str, "Error in pcap_findalldevs: %s", errbuf);
			write_event(str);
			return TRUE;
		}

		i = 0;
		for(d=alldevs; d; d=d->next)
		{			
			sprintf(str, "[%d] %s", ++i, d->description);
			SendDlgItemMessage(hDlg, IDC_COMBO_ETHIF, CB_ADDSTRING, 0, (long)str);			
		}	

		SendMessage(GetDlgItem(hDlg, IDC_COMBO_ETHIF), CB_SETCURSEL, 0, 0);

		srand( (unsigned)time( NULL ) );
		return TRUE;

	case WM_CTLCOLORSTATIC:
		if ((HWND)lParam == GetDlgItem(hDlg, IDC_STATIC_CM) || 
			(HWND)lParam == GetDlgItem(hDlg, IDC_STATIC_LT) ||
			((HWND)lParam == GetDlgItem(hDlg, IDC_EDIT_RESP)))
		{
			SetBkColor  ((HDC)wParam, WHITE);
			result = (long)hbrush1;
		}
		else if ((HWND)lParam == GetDlgItem(hDlg, IDC_STATIC_STATUS))
		{
			SetTextColor((HDC)wParam, color);
			SetBkColor  ((HDC)wParam, WHITE);
			result = (long)hbrush1;
		}
		break;

	case WM_COMMAND:
		switch (LOWORD(wParam))
		{
		case IDC_CHECK_DHCP:
			if (IsDlgButtonChecked(hDlg, IDC_CHECK_DHCP) == BST_CHECKED)
				EnableWindow(GetDlgItem(hDlg, IDC_IPADDRESS1), FALSE);
			else
				EnableWindow(GetDlgItem(hDlg, IDC_IPADDRESS1), TRUE);
			break;
		
		case IDC_INIT:
			if (adhandle)
				pcap_close(adhandle);
			
			inum = (int)SendDlgItemMessage(hDlg, IDC_COMBO_ETHIF, CB_GETCURSEL, 0, 0);
			if (inum == CB_ERR)
			{
				write_event("Local adapter not selected from list");
				break;
			}

			GetDlgItemText(hDlg, IDC_IPADDRESS1, source_ipaddr, sizeof (source_ipaddr) - 1);
			GetDlgItemText(hDlg, IDC_IPADDRESS_GW, gw_ipaddr, sizeof (gw_ipaddr) - 1);
			GetDlgItemText(hDlg, IDC_IPADDRESS_MASK, mask, sizeof (mask) - 1);
			GetDlgItemText(hDlg, IDC_IPADDRESS_DNS, dns_ipaddr, sizeof (dns_ipaddr) - 1);
			
			GetDlgItemText(hDlg, IDC_EDIT_MAC, ascii_hex_eth, sizeof (ascii_hex_eth) - 1);

			convert_hex_ethaddr_to_binary(ascii_hex_eth, ethaddr);
				
			/* Jump to the selected adapter */
			for(d=alldevs, i=0; i< inum ;d=d->next, i++);

			if ((adhandle= pcap_open_live(d->name,	// name of the device
							 65536,					// portion of the packet to capture. 
													// 65536 grants that the whole packet will be captured on all the MACs.
							 1,						// promiscuous mode (nonzero means promiscuous)
							 1,						// read timeout
							 errbuf					// error buffer
							 )) == NULL)
			{
				SetDlgItemText(dlg_main, 
					IDC_STATIC_STATUS, "Unable to open the adapter : not supported by WinPcap");
				break;
			}

			sockinit.dhcp_client = (IsDlgButtonChecked(hDlg, IDC_CHECK_DHCP) == BST_CHECKED) ? 1: 0;

			if (sockinit.dhcp_client)
				strcpy(source_ipaddr, "0.0.0.0");

			sockinit.source_ip_str = source_ipaddr;
			sockinit.gateway_ip_str = gw_ipaddr;
			sockinit.dns_server_ip_str = dns_ipaddr;
			sockinit.network_mask = mask;
			sockinit.local_eth = ethaddr;
			sockinit.timer_start_fn = sock_timer_start;
			sockinit.timer_stop_fn = sock_timer_stop;
			sockinit.drv_tx_fn = sock_drv_tx_fn;
			sockinit.log_fn = write_event;
			sockinit.global_ev_fn = hs_sock_global_event;

			sockinit.get_ticks_fn = hs_sock_get_ticks;
			sockinit.seed = (unsigned long)GetTickCount();
			sockinit.get_rand32 = hs_sock_get_rand32;

			sockinit.icmp_event_fn = icmp_event_callback;

			rc = HsSockInit(&sockinit);
			if (rc != HS_SOCK_RC_OK)
			{
				sprintf(str, "HsSockInit failed, rc=%s", HsSockGetRcString(rc));
				write_event(str);
				break;
			}

			write_event("HsTCPIPv4 Initialized");
			break;

		case IDC_SHUTDOWN:			
			HsSockCleanUp();
			
			if (adhandle)
			{
				pcap_close(adhandle);
				adhandle = NULL;
			}

			write_event("HsTCPIPv4 De-Initialized");
			break;
            
		case IDC_BUTTON_PING: 
			start_ping(hDlg);
			break;

		case IDC_BUTTON_STOP:
			stop_ping(hDlg);
			break;
		}
		
		break; 
	}
    return result;
}

/*
 * Main window message handler
 */
LRESULT CALLBACK WndProc(HWND		hWnd, 
						 UINT		message, 
						 WPARAM		wParam, 
						 LPARAM		lParam)
{
    LVITEM item = {0};

	static INITCOMMONCONTROLSEX icex;

	switch (message) 
	{
	case WM_CREATE:
		icex.dwSize = sizeof(icex);
		icex.dwICC = ICC_DATE_CLASSES;

		InitCommonControlsEx(&icex);

		InitCommonControls(); 

		memset(&fnt, 0, sizeof fnt);
		fnt.lfWidth = 8;
		fnt.lfPitchAndFamily = FF_MODERN;
		hfont	= CreateFontIndirect(&fnt);

		dlg_main = NewDialogWin(hInst, hWnd, DlgWin, IDD_DIALOG_MAIN);
		break;

	case WM_APP_TIMER:
		process_timer(lParam);
		break;

	case WM_DESTROY:
		if (hbrush1)
		{
			SelectObject(GetWindowDC(hWnd), hbrush1);
			DeleteObject(hbrush1);
		}

		if (hfont)
		{
			SelectObject(GetWindowDC(hWnd), (HGDIOBJ)hfont);
			DeleteObject(hfont);
		}

		update_config();

		HsSockCleanUp();

		PostQuitMessage(0);
		break;

	default:
		return DefWindowProc(hWnd, message, wParam, lParam);
	}
   
	return 0;
}

/*
 * Init Application Instance
 */
BOOL InitInstance(HINSTANCE hInstance, int nCmdShow, HWND *hwnd)
{
	HWND hWnd;

	hInst = hInstance;

	hWnd = CreateWindow(szWindowClass, 
					   szTitle, 
					   WS_OVERLAPPED | WS_SYSMENU,
					   0, 
					   0, 
					   MAINWIDTH,
					   MAINDEPTH, 
					   NULL, LoadMenu(hInstance, 0/*(LPCTSTR)IDR_MENU1*/), 
					   hInstance, NULL);

	if (!hWnd)
	{
		return FALSE;
	}
	*hwnd = hWnd;

	ShowWindow(hWnd, nCmdShow);
	UpdateWindow(hWnd);

	return TRUE;
}

/*
 * entry point
 */
int APIENTRY WinMain(HINSTANCE hInstance,
                     HINSTANCE hPrevInstance,
                     LPSTR     lpCmdLine,
                     int       nCmdShow)
{
	MSG		msg;
	HACCEL	hAccelTable;
	unsigned char *pkt_data;
	struct pcap_pkthdr *header;
	MMRESULT res;
	int r;

	res = timeGetDevCaps(&timecaps, sizeof (TIMECAPS));
	if (res == TIMERR_NOERROR)
	{			
		timeBeginPeriod((UINT)timecaps.wPeriodMin);
	}

	hbrush1 = CreateSolidBrush(RGB(255, 255, 255));

	get_config();

	BlRegisterClass(hInstance, CS_HREDRAW | CS_VREDRAW, WndProc, (int)IDC_ARROW,
				   (HBRUSH)(COLOR_APPWORKSPACE+1), szWindowClass, IDI_ICON1, 0);

	if (!InitInstance (hInstance, nCmdShow, &hwnd_main)) 
	{
		return FALSE;
	}

	hAccelTable = LoadAccelerators(hInstance, (LPCTSTR)0);

	while (1) 
	{
		if (PeekMessage(&msg, NULL, 0, 0, PM_NOREMOVE))
		{
			if (GetMessage(&msg, NULL, 0, 0))
			{
				if (!TranslateAccelerator(msg.hwnd, hAccelTable, &msg)) 
				{
					TranslateMessage(&msg);
					DispatchMessage(&msg);
				}
			}
			else
			{
				break;
			}
		}
		else
		{
			Sleep(1);

			if (adhandle)
			{
				r = pcap_next_ex( adhandle, &header, &pkt_data);
				if (r > 0)
				{
					HsIpReceivePacket(pkt_data, (int)header->caplen);					
				}
			}			
		}		
	}

	if (alldevs)
		pcap_freealldevs(alldevs);

	if (adhandle)
		pcap_close(adhandle);

	return msg.wParam;	
}


